!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!    #################### 
     MODULE MODI_TURB_VER_THERMO_FLUX 
!    ####################
!
INTERFACE 
!
      SUBROUTINE TURB_VER_THERMO_FLUX(KKA,KKU,KKL,KRR,KRRL,KRRI,    &
                      OTURB_FLX,HTURBDIM,HTOM,                      &
                      PIMPL,PEXPL,                                  &
                      PTSTEP,                                       &
                      TPFILE,                                       &
                      PDZZ,PDIRCOSZW,PZZ,                           &
                      PRHODJ,                                       &
                      PSFTHM,PSFRM,PSFTHP,PSFRP,                    &
                      PWM,PTHLM,PRM,PSVM,                           &
                      PTKEM,PLM,PLEPS,                              &
                      PATHETA,PAMOIST,PSRCM,PFRAC_ICE,              &
                      PBETA, PSQRT_TKE, PDTH_DZ, PDR_DZ, PRED2TH3,  &
                      PRED2R3, PRED2THR3, PBLL_O_E, PETHETA,        &
                      PEMOIST, PREDTH1, PREDR1, PPHI3, PPSI3, PD,   &
                      PFWTH,PFWR,PFTH2,PFR2,PFTHR,PBL_DEPTH,        &
                      PWTHV,PRTHLS,PRRS,PTHLP,PRP,PTP,PWTH,PWRC     )
!
USE MODD_IO, ONLY: TFILEDATA
!
INTEGER,                INTENT(IN)   :: KKA           !near ground array index  
INTEGER,                INTENT(IN)   :: KKU           !uppest atmosphere array index
INTEGER,                INTENT(IN)   :: KKL           !vert. levels type 1=MNH -1=AR O
INTEGER,                INTENT(IN)   :: KRR           ! number of moist var.
INTEGER,                INTENT(IN)   :: KRRL          ! number of liquid water var.
INTEGER,                INTENT(IN)   :: KRRI          ! number of ice water var.
LOGICAL,                INTENT(IN)   ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
CHARACTER(len=4),       INTENT(IN)   ::  HTURBDIM     ! dimensionality of the
                                                      ! turbulence scheme
CHARACTER(len=4),       INTENT(IN)   ::  HTOM         ! type of Third Order Moment
REAL,                   INTENT(IN)   ::  PIMPL, PEXPL ! Coef. for temporal disc.
REAL,                   INTENT(IN)   ::  PTSTEP       ! Double Time Step
TYPE(TFILEDATA),        INTENT(INOUT)::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDZZ         ! Metric coefficients
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PDIRCOSZW    ! Director Cosinus of the
                                                      ! normal to the ground surface
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PZZ          ! altitudes
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRHODJ       ! dry density * grid volum
!
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSFTHM,PSFRM ! surface fluxes at time
!                                                     ! t - deltat 
!
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSFTHP,PSFRP ! surface fluxes at time
!                                                     ! t + deltat 
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PWM 
! Vertical wind
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTHLM 
! potential temperature at t-Delta t
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PRM          ! Mixing ratios 
                                                      ! at t-Delta t
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PSVM         ! Mixing ratios 
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTKEM        ! TKE at time t
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLM          ! Turb. mixing length   
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLEPS        ! dissipative length   
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PATHETA      ! coefficients between 
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PAMOIST      ! s and Thetal and Rnp
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PSRCM        ! normalized 
                  ! 2nd-order flux s'r'c/2Sigma_s2 at t-1 multiplied by Lambda_3
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFRAC_ICE    ! ri fraction of rc+ri
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PBETA        ! buoyancy coefficient
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PSQRT_TKE    ! sqrt(e)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDTH_DZ      ! d(th)/dz
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDR_DZ       ! d(rt)/dz
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2TH3     ! 3D Redeslperger number R*2_th
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2R3      ! 3D Redeslperger number R*2_r
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2THR3    ! 3D Redeslperger number R*2_thr
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PBLL_O_E     ! beta * Lk * Leps / tke
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PETHETA      ! Coefficient for theta in theta_v computation
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PEMOIST      ! Coefficient for r in theta_v computation
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PREDTH1      ! 1D Redelsperger number for Th
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PREDR1       ! 1D Redelsperger number for r
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PPHI3        ! Prandtl number for temperature
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PPSI3        ! Prandtl number for vapor
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PD           ! Denominator in Prandtl numbers
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFWTH        ! d(w'2th' )/dz (at flux point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFWR         ! d(w'2r'  )/dz (at flux point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFTH2        ! d(w'th'2 )/dz (at mass point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFR2         ! d(w'r'2  )/dz (at mass point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFTHR        ! d(w'th'r')/dz (at mass point)
REAL, DIMENSION(:,:),   INTENT(INOUT)::  PBL_DEPTH    ! BL depth
REAL, DIMENSION(:,:,:), INTENT(OUT)  :: PWTHV         ! buoyancy flux
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRTHLS     ! cumulated source for theta
REAL, DIMENSION(:,:,:,:), INTENT(INOUT) :: PRRS       ! cumulated source for rt
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PTHLP      ! guess of thl at t+ deltat
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PRP        ! guess of r at t+ deltat
!
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PTP       ! Dynamic and thermal
                                                     ! TKE production terms
!
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PWTH       ! heat flux
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PWRC       ! cloud water flux
!
END SUBROUTINE TURB_VER_THERMO_FLUX
!
END INTERFACE
!
END MODULE MODI_TURB_VER_THERMO_FLUX
!
!
!     ###############################################################
      SUBROUTINE TURB_VER_THERMO_FLUX(KKA,KKU,KKL,KRR, KRRL, KRRI,  &
                      OTURB_FLX,HTURBDIM,HTOM,                      &
                      PIMPL,PEXPL,                                  &
                      PTSTEP,                                       &
                      TPFILE,                                       &
                      PDZZ,PDIRCOSZW,PZZ,                           &
                      PRHODJ,                                       &
                      PSFTHM,PSFRM,PSFTHP,PSFRP,                    &
                      PWM,PTHLM,PRM,PSVM,                           &
                      PTKEM,PLM,PLEPS,                              &
                      PATHETA,PAMOIST,PSRCM,PFRAC_ICE,              &
                      PBETA, PSQRT_TKE, PDTH_DZ, PDR_DZ, PRED2TH3,  &
                      PRED2R3, PRED2THR3, PBLL_O_E, PETHETA,        &
                      PEMOIST, PREDTH1, PREDR1, PPHI3, PPSI3, PD,   &
                      PFWTH,PFWR,PFTH2,PFR2,PFTHR,PBL_DEPTH,        &
                      PWTHV,PRTHLS,PRRS,PTHLP,PRP,PTP,PWTH,PWRC     )
!     ###############################################################
!
!
!!****  *TURB_VER_THERMO_FLUX* -compute the source terms due to the vertical turbulent
!!       fluxes.
!!
!!    PURPOSE
!!    -------
!       The purpose of this routine is to compute the vertical turbulent
!     fluxes of the evolutive variables and give back the source 
!     terms to the main program.	In the case of large horizontal meshes,
!     the divergence of these vertical turbulent fluxes represent the whole
!     effect of the turbulence but when the three-dimensionnal version of
!     the turbulence scheme is activated (CTURBDIM="3DIM"), these divergences
!     are completed in the next routine TURB_HOR. 
!		  An arbitrary degree of implicitness has been implemented for the 
!     temporal treatment of these diffusion terms.
!       The vertical boundary conditions are as follows:
!           *  at the bottom, the surface fluxes are prescribed at the same
!              as the other turbulent fluxes
!           *  at the top, the turbulent fluxes are set to 0.
!       It should be noted that the condensation has been implicitely included
!     in this turbulence scheme by using conservative variables and computing
!     the subgrid variance of a statistical variable s indicating the presence 
!     or not of condensation in a given mesh. 
!
!!**  METHOD
!!    ------
!!      1D type calculations are made;
!!      The vertical turbulent fluxes are computed in an off-centered
!!      implicit scheme (a Crank-Nicholson type with coefficients different
!!      than 0.5), which allows to vary the degree of implicitness of the
!!      formulation.
!!      	 The different prognostic variables are treated one by one. 
!!      The contributions of each turbulent fluxes are cumulated into the 
!!      tendency  PRvarS, and into the dynamic and thermal production of 
!!      TKE if necessary.
!!        
!!			 In section 2 and 3, the thermodynamical fields are considered.
!!      Only the turbulent fluxes of the conservative variables
!!      (Thetal and Rnp stored in PRx(:,:,:,1))  are computed. 
!!       Note that the turbulent fluxes at the vertical 
!!      boundaries are given either by the soil scheme for the surface one
!!      ( at the same instant as the others fluxes) and equal to 0 at the 
!!      top of the model. The thermal production is computed by vertically 
!!      averaging the turbulent flux and multiply this flux at the mass point by
!!      a function ETHETA or EMOIST, which preform the transformation from the
!!      conservative variables to the virtual potential temperature. 
!!     
!! 	    In section 4, the variance of the statistical variable
!!      s indicating presence or not of condensation, is determined in function 
!!      of the turbulent moments of the conservative variables and its
!!      squarred root is stored in PSIGS. This information will be completed in 
!!      the horizontal turbulence if the turbulence dimensionality is not 
!!      equal to "1DIM".
!!
!!			 In section 5, the x component of the stress tensor is computed.
!!      The surface flux <u'w'> is computed from the value of the surface
!!      fluxes computed in axes linked to the orography ( i", j" , k"):
!!        i" is parallel to the surface and in the direction of the maximum
!!           slope
!!        j" is also parallel to the surface and in the normal direction of
!!           the maximum slope
!!        k" is the normal to the surface
!!      In order to prevent numerical instability, the implicit scheme has 
!!      been extended to the surface flux regarding to its dependence in 
!!      function of U. The dependence in function of the other components 
!!      introduced by the different rotations is only explicit.
!!      The turbulent fluxes are used to compute the dynamic production of 
!!      TKE. For the last TKE level ( located at PDZZ(:,:,IKB)/2 from the
!!      ground), an harmonic extrapolation from the dynamic production at 
!!      PDZZ(:,:,IKB) is used to avoid an evaluation of the gradient of U
!!      in the surface layer.
!!
!!         In section 6, the same steps are repeated but for the y direction
!!		  and in section 7, a diagnostic computation of the W variance is 
!!      performed.
!!
!!         In section 8, the turbulent fluxes for the scalar variables are 
!!      computed by the same way as the conservative thermodynamical variables
!!
!!            
!!    EXTERNAL
!!    --------
!!      GX_U_M, GY_V_M, GZ_W_M :  cartesian gradient operators 
!!      GX_U_UW,GY_V_VW	         (X,Y,Z) represent the direction of the gradient
!!                               _(M,U,...)_ represent the localization of the 
!!                               field to be derivated
!!                               _(M,UW,...) represent the localization of the 
!!                               field	derivated
!!                               
!!
!!      MXM,MXF,MYM,MYF,MZM,MZF
!!                             :  Shuman functions (mean operators)     
!!      DXF,DYF,DZF,DZM
!!                             :  Shuman functions (difference operators)     
!!                               
!!      SUBROUTINE TRIDIAG     : to compute the split implicit evolution
!!                               of a variable located at a mass point
!!
!!      SUBROUTINE TRIDIAG_WIND: to compute the split implicit evolution
!!                               of a variable located at a wind point
!!
!!      FUNCTIONs ETHETA and EMOIST  :  
!!            allows to compute:
!!            - the coefficients for the turbulent correlation between
!!            any variable and the virtual potential temperature, of its 
!!            correlations with the conservative potential temperature and 
!!            the humidity conservative variable:
!!            -------              -------              -------
!!            A' Thv'  =  ETHETA   A' Thl'  +  EMOIST   A' Rnp'  
!!
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!      Module MODD_CST : contains physical constants
!!
!!           XG         : gravity constant
!!
!!      Module MODD_CTURB: contains the set of constants for
!!                        the turbulence scheme
!!
!!           XCMFS,XCMFB : cts for the momentum flux
!!           XCSHF       : ct for the sensible heat flux
!!           XCHF        : ct for the moisture flux
!!           XCTV,XCHV   : cts for the T and moisture variances
!!
!!      Module MODD_PARAMETERS
!!
!!           JPVEXT_TURB     : number of vertical external points
!!           JPHEXT     : number of horizontal external points
!!
!!
!!    REFERENCE
!!    ---------
!!      Book 1 of documentation (Chapter: Turbulence)
!!
!!    AUTHOR
!!    ------
!!      Joan Cuxart             * INM and Meteo-France *
!!
!!    MODIFICATIONS
!!    -------------
!!      Original       August   19, 1994
!!      Modifications: February 14, 1995 (J.Cuxart and J.Stein) 
!!                                  Doctorization and Optimization
!!      Modifications: March 21, 1995 (J.M. Carriere) 
!!                                  Introduction of cloud water
!!      Modifications: June  14, 1995 (J.Cuxart and J. Stein) 
!!                                 Phi3 and Psi3 at w-point + bug in the all
!!                                 or nothing condens. 
!!      Modifications: Sept  15, 1995 (J.Cuxart and J. Stein) 
!!                                 Change the DP computation at the ground
!!      Modifications: October 10, 1995 (J.Cuxart and J. Stein) 
!!                                 Psi for scal var and LES tools
!!      Modifications: November 10, 1995 (J. Stein)
!!                                 change the surface	relations 
!!      Modifications: February 20, 1995 (J. Stein) optimization
!!      Modifications: May 21, 1996 (J. Stein) 
!!                                  bug in the vertical flux of the V wind 
!!                                  component for explicit computation
!!      Modifications: May 21, 1996 (N. wood) 
!!                                  modify the computation of the vertical
!!                                   part or the surface tangential flux
!!      Modifications: May 21, 1996 (P. Jabouille)
!!                                  same modification in the Y direction
!!      
!!      Modifications: Sept 17, 1996 (J. Stein) change the moist case by using
!!                                  Pi instead of Piref + use Atheta and Amoist
!!
!!      Modifications: Nov  24, 1997 (V. Masson) removes the DO loops 
!!      Modifications: Mar  31, 1998 (V. Masson) splits the routine TURB_VER_THERMO_FLUX 
!!      Modifications: Oct  18, 2000 (V. Masson) LES computations
!!      Modifications: Dec  01, 2000 (V. Masson) conservation of energy from
!!                                               surface flux in 1DIM case
!!                                               when slopes are present
!!                     Nov  06, 2002 (V. Masson) LES budgets
!!                     Feb  20, 2003 (JP Pinty)  Add PFRAC_ICE
!!                     May  20, 2003 (JP Pinty)  Correction of ETHETA
!!                                                         and EMOIST calls
!!                     July     2005 (S. Tomas, V. Masson)
!!                                               Add 3rd order moments
!!                                               and implicitation of PHI3 and PSI3
!!                     October 2009 (G. Tanguy) add ILENCH=LEN(YCOMMENT) after
!!                                              change of YCOMMENT
!!                     2012-02 (Y. Seity) add possibility to run with reversed
!!                                             vertical levels
!!                     04/2016 (M.Moge) Use openACC directives to port the TURB part of Meso-NH on GPU
!!  Philippe Wautelet: 05/2016-04/2018: new data structures and calls for I/O
!!                     2021 (D. Ricard) last version of HGRAD turbulence scheme
!!                                 Leronard terms instead of Reynolds terms
!!                                 applied to vertical fluxes of r_np and Thl
!!                                 for implicit version of turbulence scheme
!!                                 corrections and cleaning
!!                     June 2020 (B. Vie) Patch preventing negative rc and ri in 2.3 and 3.3
!! JL Redelsperger  : 03/2021: Ocean and Autocoupling O-A LES Cases
!!                             Sfc flux shape for LDEEPOC Case
!!--------------------------------------------------------------------------
!       
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST
USE MODD_CTURB
use modd_field,          only: tfielddata, TYPEREAL
USE MODD_GRID_n,         ONLY: XZS, XXHAT, XYHAT
USE MODD_IO,             ONLY: TFILEDATA
USE MODD_METRICS_n,      ONLY: XDXX, XDYY, XDZX, XDZY, XDZZ
USE MODD_PARAMETERS
USE MODD_TURB_n,         ONLY: LHGRAD, XCOEFHGRADTHL, XCOEFHGRADRM, XALTHGRAD, XCLDTHOLD
USE MODD_CONF
USE MODD_LES
USE MODD_DIM_n
USE MODD_DYN_n,          ONLY: LOCEAN
USE MODD_OCEANH
USE MODD_REF,            ONLY: LCOUPLES
USE MODD_TURB_n
USE MODD_FRC
!
USE MODI_GRADIENT_U
USE MODI_GRADIENT_V
USE MODI_GRADIENT_W
USE MODI_GRADIENT_M
#ifndef MNH_OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN !Kept because some parts are not yet ported with OpenACC
USE MODI_SHUMAN_DEVICE
#endif
USE MODI_TRIDIAG 
USE MODI_LES_MEAN_SUBGRID
USE MODI_TRIDIAG_THERMO
USE MODI_TM06_H
!
USE MODE_GATHER_ll
USE MODE_IO_FIELD_WRITE, only: IO_Field_write
USE MODE_ll
USE MODE_MPPDB
#ifdef MNH_OPENACC
USE MODE_MSG
#endif
USE MODE_PRANDTL
!
USE MODI_SECOND_MNH
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif

IMPLICIT NONE
!
!*      0.1  declarations of arguments
!
!
!
INTEGER,                INTENT(IN)   :: KKA           !near ground array index  
INTEGER,                INTENT(IN)   :: KKU           !uppest atmosphere array index
INTEGER,                INTENT(IN)   :: KKL           !vert. levels type 1=MNH -1=ARO
INTEGER,                INTENT(IN)   :: KRR           ! number of moist var.
INTEGER,                INTENT(IN)   :: KRRL          ! number of liquid water var.
INTEGER,                INTENT(IN)   :: KRRI          ! number of ice water var.
LOGICAL,                INTENT(IN)   ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
CHARACTER(len=4),       INTENT(IN)   ::  HTURBDIM     ! dimensionality of the
                                                      ! turbulence scheme
CHARACTER(len=4),       INTENT(IN)   ::  HTOM         ! type of Third Order Moment
REAL,                   INTENT(IN)   ::  PIMPL, PEXPL ! Coef. for temporal disc.
REAL,                   INTENT(IN)   ::  PTSTEP       ! Double Time Step
TYPE(TFILEDATA),        INTENT(INOUT)::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDZZ         ! Metric coefficients
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PDIRCOSZW    ! Director Cosinus of the
                                                      ! normal to the ground surface
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PZZ          ! altitudes
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRHODJ       ! dry density * grid volum
!
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSFTHM,PSFRM ! surface fluxes at time
!                                                     ! t - deltat 
!
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSFTHP,PSFRP ! surface fluxes at time
!                                                     ! t + deltat 
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PWM 
! Vertical wind
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTHLM 
! potential temperature at t-Delta t
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PRM          ! Mixing ratios 
                                                      ! at t-Delta t
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PSVM         ! Mixing ratios 
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTKEM        ! TKE at time t
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLM          ! Turb. mixing length   
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLEPS        ! dissipative length   
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PATHETA      ! coefficients between 
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PAMOIST      ! s and Thetal and Rnp
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PSRCM        ! normalized 
! 2nd-order flux s'r'c/2Sigma_s2 at t-1 multiplied by Lambda_3
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFRAC_ICE    ! ri fraction of rc+ri
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PBETA        ! buoyancy coefficient
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PSQRT_TKE    ! sqrt(e)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDTH_DZ      ! d(th)/dz
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDR_DZ       ! d(rt)/dz
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2TH3     ! 3D Redeslperger number R*2_th
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2R3      ! 3D Redeslperger number R*2_r
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2THR3    ! 3D Redeslperger number R*2_thr
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PBLL_O_E     ! beta * Lk * Leps / tke
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PETHETA      ! Coefficient for theta in theta_v computation
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PEMOIST      ! Coefficient for r in theta_v computation
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PREDTH1      ! 1D Redelsperger number for Th
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PREDR1       ! 1D Redelsperger number for r
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PPHI3        ! Prandtl number for temperature
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PPSI3        ! Prandtl number for vapor
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PD           ! Denominator in Prandtl numbers
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFWTH        ! d(w'2th' )/dz (at flux point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFWR         ! d(w'2r'  )/dz (at flux point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFTH2        ! d(w'th'2 )/dz (at mass point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFR2         ! d(w'r'2  )/dz (at mass point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFTHR        ! d(w'th'r')/dz (at mass point)
REAL, DIMENSION(:,:),   INTENT(INOUT)::  PBL_DEPTH    ! BL depth
REAL, DIMENSION(:,:,:), INTENT(OUT)  :: PWTHV         ! buoyancy flux
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRTHLS     ! cumulated source for theta
REAL, DIMENSION(:,:,:,:), INTENT(INOUT) :: PRRS       ! cumulated source for rt
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PTHLP      ! guess of thl at t+ deltat
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PRP        ! guess of r at t+ deltat
!
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PTP       ! Dynamic and thermal
                                                     ! TKE production terms
!
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PWTH       ! heat flux
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PWRC       ! cloud water flux
!
!*       0.2  declaration of local variables
!
INTEGER             :: IKB,IKE      ! I index values for the Beginning and End
                                    ! mass points of the domain in the 3 direct.
INTEGER             :: IKT          ! array size in k direction
INTEGER             :: IKTB,IKTE    ! start, end of k loops in physical domain
!
INTEGER             :: JI, JJ ! loop indexes 
!
INTEGER                    :: IIB,IJB       ! Lower bounds of the physical
                                            ! sub-domain in x and y directions
INTEGER                    :: IIE,IJE       ! Upper bounds of the physical
                                            ! sub-domain in x and y directions
!
REAL, DIMENSION(:), ALLOCATABLE   :: ZXHAT_ll    !  Position x in the conformal
                                                 ! plane (array on the complete domain)
REAL, DIMENSION(:), ALLOCATABLE   :: ZYHAT_ll    !   Position y in the conformal
                                                 ! plane (array on the complete domain)
!
!
CHARACTER (LEN=100) :: YCOMMENT     ! comment string in LFIFM file
CHARACTER (LEN=LEN_HREC)  :: YRECFM       ! Name of the desired field in LFIFM file
!
REAL :: ZTIME1, ZTIME2
REAL :: ZDELTAX
REAL    :: ZXBEG,ZXEND,ZYBEG,ZYEND ! Forcing size for ocean deep convection
REAL, DIMENSION(SIZE(XXHAT),SIZE(XYHAT)) :: ZDIST ! distance
                                   ! from the center of the cooling               
REAL :: ZFLPROV
INTEGER           :: JKM          ! vertical index loop
INTEGER          :: JSW
REAL :: ZSWA     ! index for time flux interpolation
!
INTEGER :: IIU, IJU
INTEGER :: IRESP
INTEGER :: JK
LOGICAL :: GUSERV   ! flag to use water
LOGICAL :: GFTH2    ! flag to use w'th'2
LOGICAL :: GFWTH    ! flag to use w'2th'
LOGICAL :: GFR2     ! flag to use w'r'2
LOGICAL :: GFWR     ! flag to use w'2r'
LOGICAL :: GFTHR    ! flag to use w'th'r'
!
REAL, DIMENSION(:,:,:), pointer , contiguous ::  &
       ZA,         & ! work variable for wrc or LES computation
       ZFLXZ,      & ! vertical flux of the treated variable
       ZSOURCE,    & ! source of evolution for the treated variable
       ZKEFF,      & ! effectif diffusion coeff = LT * SQRT( TKE )
       ZF,         & ! Flux in dTh/dt =-dF/dz (evaluated at t-1)(or rt instead of Th)
       ZDFDDTDZ,   & ! dF/d(dTh/dz)
       ZDFDDRDZ,   & ! dF/d(dr/dz)
       Z3RDMOMENT, & ! 3 order term in flux or variance equation
       ZF_NEW,     &
       ZRWTHL,     &
       ZRWRNP,     &
       ZCLD_THOLD
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZALT
#ifdef MNH_OPENACC
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP1_DEVICE,ZTMP2_DEVICE,ZTMP3_DEVICE,ZTMP4_DEVICE
#endif
TYPE(TFIELDDATA) :: TZFIELD

INTEGER :: JIU,JJU,JKU
LOGICAL :: GOCEAN !Intermediate variable used to work around a Cray compiler bug (CCE 13.0.0)
REAL :: ZALTHGRAD, ZCLDTHOLD !Intermediate variable used to work around a Cray compiler bug (CCE 13.0.0)

!----------------------------------------------------------------------------

!$acc data present( PDZZ, PDIRCOSZW, PZZ,                            &
!$acc &             PRHODJ,                                          &
!$acc &             PSFTHM, PSFRM, PSFTHP, PSFRP,                    &
!$acc &             PWM, PTHLM, PRM, PSVM,                           &
!$acc &             PTKEM, PLM, PLEPS,                               &
!$acc &             PATHETA, PAMOIST, PSRCM, PFRAC_ICE,              &
!$acc &             PBETA,  PSQRT_TKE,  PDTH_DZ,  PDR_DZ,  PRED2TH3, &
!$acc &             PRED2R3,  PRED2THR3,  PBLL_O_E,  PETHETA,        &
!$acc &             PEMOIST,  PREDTH1,  PREDR1,  PPHI3,  PPSI3,  PD, &
!$acc &             PFWTH, PFWR, PFTH2, PFR2, PFTHR, PBL_DEPTH,      &
!$acc &             PWTHV, PRTHLS, PRRS, PTHLP, PRP, PTP, PWTH, PWRC )

if ( mppdb_initialized ) then
  !check all in arrays
  call Mppdb_check( pdzz,      "Turb_ver_thermo_flux beg:pdzz"      )
  call Mppdb_check( pdircoszw, "Turb_ver_thermo_flux beg:pdircoszw" )
  call Mppdb_check( pzz,       "Turb_ver_thermo_flux beg:pzz"       )
  call Mppdb_check( prhodj,    "Turb_ver_thermo_flux beg:prhodj"    )
  call Mppdb_check( psfthm,    "Turb_ver_thermo_flux beg:psfthm"    )
  call Mppdb_check( psfrm,     "Turb_ver_thermo_flux beg:psfrm"     )
  call Mppdb_check( psfthp,    "Turb_ver_thermo_flux beg:psfthp"    )
  call Mppdb_check( psfrp,     "Turb_ver_thermo_flux beg:psfrp"     )
  call Mppdb_check( pwm,       "Turb_ver_thermo_flux beg:pwm"       )
  call Mppdb_check( pthlm,     "Turb_ver_thermo_flux beg:pthlm"     )
  call Mppdb_check( prm,       "Turb_ver_thermo_flux beg:prm"       )
  call Mppdb_check( psvm,      "Turb_ver_thermo_flux beg:psvm"      )
  call Mppdb_check( ptkem,     "Turb_ver_thermo_flux beg:ptkem"     )
  call Mppdb_check( plm,       "Turb_ver_thermo_flux beg:plm"       )
  call Mppdb_check( pleps,     "Turb_ver_thermo_flux beg:pleps"     )
  call Mppdb_check( patheta,   "Turb_ver_thermo_flux beg:patheta"   )
  call Mppdb_check( pamoist,   "Turb_ver_thermo_flux beg:pamoist"   )
  call Mppdb_check( psrcm,     "Turb_ver_thermo_flux beg:psrcm"     )
  call Mppdb_check( pfrac_ice, "Turb_ver_thermo_flux beg:pfrac_ice" )
  call Mppdb_check( pbeta,     "Turb_ver_thermo_flux beg:pbeta"     )
  call Mppdb_check( psqrt_tke, "Turb_ver_thermo_flux beg:psqrt_tke" )
  call Mppdb_check( pdth_dz,   "Turb_ver_thermo_flux beg:pdth_dz"   )
  call Mppdb_check( pdr_dz,    "Turb_ver_thermo_flux beg:pdr_dz"    )
  call Mppdb_check( pred2th3,  "Turb_ver_thermo_flux beg:pred2th3"  )
  call Mppdb_check( pred2r3,   "Turb_ver_thermo_flux beg:pred2r3"   )
  call Mppdb_check( pred2thr3, "Turb_ver_thermo_flux beg:pred2thr3" )
  call Mppdb_check( pbll_o_e,  "Turb_ver_thermo_flux beg:pbll_o_e"  )
  call Mppdb_check( petheta,   "Turb_ver_thermo_flux beg:petheta"   )
  call Mppdb_check( pemoist,   "Turb_ver_thermo_flux beg:pemoist"   )
  call Mppdb_check( predth1,   "Turb_ver_thermo_flux beg:predth1"   )
  call Mppdb_check( predr1,    "Turb_ver_thermo_flux beg:predr1"    )
  call Mppdb_check( pphi3,     "Turb_ver_thermo_flux beg:pphi3"     )
  call Mppdb_check( ppsi3,     "Turb_ver_thermo_flux beg:ppsi3"     )
  call Mppdb_check( pd,        "Turb_ver_thermo_flux beg:pd"        )
  call Mppdb_check( pfwth,     "Turb_ver_thermo_flux beg:pfwth"     )
  call Mppdb_check( pfwr,      "Turb_ver_thermo_flux beg:pfwr"      )
  call Mppdb_check( pfth2,     "Turb_ver_thermo_flux beg:pfth2"     )
  call Mppdb_check( pfr2,      "Turb_ver_thermo_flux beg:pfr2"      )
  call Mppdb_check( pfthr,     "Turb_ver_thermo_flux beg:pfthr"     )
  !check all inout arrays
  call Mppdb_check( pbl_depth, "Turb_ver_thermo_flux beg:pbl_depth" )
  call Mppdb_check( prthls,    "Turb_ver_thermo_flux beg:prthls"    )
  call Mppdb_check( prrs,      "Turb_ver_thermo_flux beg:prrs"      )
end if

JIU =  size( pthlm, 1 )
JJU =  size( pthlm, 2 )
JKU =  size( pthlm, 3 )

GOCEAN = LOCEAN
ZALTHGRAD = XALTHGRAD
ZCLDTHOLD = XCLDTHOLD

#ifndef MNH_OPENACC
allocate( za        (JIU,JJU,JKU ) )
allocate( zflxz     (JIU,JJU,JKU ) )
allocate( zsource   (JIU,JJU,JKU ) )
allocate( zkeff     (JIU,JJU,JKU ) )
allocate( zf        (JIU,JJU,JKU ) )
allocate( zdfddtdz  (JIU,JJU,JKU ) )
allocate( zdfddrdz  (JIU,JJU,JKU ) )
allocate( z3rdmoment(JIU,JJU,JKU ) )
allocate( zf_new    (JIU,JJU,JKU ) )
allocate( zrwthl    (JIU,JJU,JKU ) )
allocate( zrwrnp    (JIU,JJU,JKU ) )
allocate( zcld_thold(JIU,JJU,JKU ) )
allocate( zalt      (JIU,JJU,JKU ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN()

CALL MNH_MEM_GET( za,         JIU,JJU,JKU )
CALL MNH_MEM_GET( zflxz,      JIU,JJU,JKU )
CALL MNH_MEM_GET( zsource,    JIU,JJU,JKU )
CALL MNH_MEM_GET( zkeff,      JIU,JJU,JKU )
CALL MNH_MEM_GET( zf,         JIU,JJU,JKU )
CALL MNH_MEM_GET( zdfddtdz,   JIU,JJU,JKU )
CALL MNH_MEM_GET( zdfddrdz,   JIU,JJU,JKU )
CALL MNH_MEM_GET( z3rdmoment, JIU,JJU,JKU )
CALL MNH_MEM_GET( zf_new,     JIU,JJU,JKU )
CALL MNH_MEM_GET( zrwthl,     JIU,JJU,JKU )
CALL MNH_MEM_GET( zrwrnp,     JIU,JJU,JKU )
CALL MNH_MEM_GET( zcld_thold, JIU,JJU,JKU )
CALL MNH_MEM_GET( zalt,       JIU,JJU,JKU )

CALL MNH_MEM_GET( ztmp1_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp2_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp3_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp4_device, JIU, JJU, JKU )
#endif

!$acc data present( ZA, ZFLXZ, ZSOURCE, ZKEFF, ZF, ZDFDDTDZ, ZDFDDRDZ, Z3RDMOMENT,  &
!$acc &             ZF_NEW, ZRWTHL, ZRWRNP, ZCLD_THOLD, ZALT,                       &
!$acc &             ZTMP1_DEVICE, ZTMP2_DEVICE, ZTMP3_DEVICE, ZTMP4_DEVICE )

!
!*       1.   PRELIMINARIES
!             -------------
! Size for a given proc & a given model      
IIU=SIZE(PTHLM,1) 
IJU=SIZE(PTHLM,2)
!
!! Compute Shape of sfc flux for Oceanic Deep Conv Case
! 
IF (GOCEAN .AND. LDEEPOC) THEN
#ifdef MNH_OPENACC
  call Print_msg( NVERB_FATAL, 'GEN', 'TURB_VER_THERMO_FLUX', 'OpenACC: LOCEAN=T not yet implemented' )
#endif
  !*       COMPUTES THE PHYSICAL SUBDOMAIN BOUNDS
  ALLOCATE(ZXHAT_ll(NIMAX_ll+2*JPHEXT),ZYHAT_ll(NJMAX_ll+2*JPHEXT))
  !compute ZXHAT_ll = position in the (0:Lx) domain 1 (Lx=Size of domain1 )
  !compute XXHAT_ll = position in the (L0_subproc,Lx_subproc) domain for the current subproc
  !                                     L0_subproc as referenced in the full domain 1
  CALL GATHERALL_FIELD_ll('XX',XXHAT,ZXHAT_ll,IRESP)
  CALL GATHERALL_FIELD_ll('YY',XYHAT,ZYHAT_ll,IRESP)
  CALL GET_DIM_EXT_ll('B',IIU,IJU)
  CALL GET_INDICE_ll(IIB,IJB,IIE,IJE)
  DO JJ = IJB,IJE
    DO JI = IIB,IIE
      ZDIST(JI,JJ) = SQRT(                         &
      (( (XXHAT(JI)+XXHAT(JI+1))*0.5 - XCENTX_OC ) / XRADX_OC)**2 + &
      (( (XYHAT(JJ)+XYHAT(JJ+1))*0.5 - XCENTY_OC ) / XRADY_OC)**2   &
                                )
    END DO
  END DO
  DO JJ=IJB,IJE
    DO JI=IIB,IIE
      IF ( ZDIST(JI,JJ) > 1.) XSSTFL(JI,JJ)=0.
    END DO
  END DO
END IF !END DEEP OCEAN CONV CASE
!
IKT  =SIZE(PTHLM,3)  
IKTE =IKT-JPVEXT_TURB  
IKTB =1+JPVEXT_TURB               
IKB=KKA+JPVEXT_TURB*KKL
IKE=KKU-JPVEXT_TURB*KKL
!
GUSERV = (KRR/=0)
!
!  compute the coefficients for the uncentred gradient computation near the 
!  ground
!
#ifndef MNH_OPENACC
ZKEFF(:,:,:) = MZM( PLM(:,:,:) * SQRT(PTKEM(:,:,:)) )
#else
!$acc kernels
ZTMP1_DEVICE(:,:,:) = PLM(:,:,:) * SQRT(PTKEM(:,:,:))
!$acc end kernels
CALL MZM_DEVICE(ZTMP1_DEVICE, ZKEFF)
#endif
!
! define a cloud mask with ri and rc (used after with a threshold) for Leonard terms
!
IF(LHGRAD) THEN
#ifdef MNH_OPENACC
  call Print_msg( NVERB_FATAL, 'GEN', 'TURB_VER_THERMO_FLUX', 'OpenACC: LHGRAD=T not yet implemented' )
#endif
!$acc kernels present_cr(zcld_thold)
  IF ( KRRL >= 1 ) THEN
    IF ( KRRI >= 1 ) THEN
      ZCLD_THOLD(:,:,:) = PRM(:,:,:,2) + PRM(:,:,:,4)
    ELSE
      ZCLD_THOLD(:,:,:) = PRM(:,:,:,2)
    END IF
  END IF
!$acc end kernels
END IF
!
! Flags for 3rd order quantities
!
GFTH2 = .FALSE.
GFR2  = .FALSE.
GFTHR = .FALSE.
GFWTH = .FALSE.
GFWR  = .FALSE.
!
IF (HTOM/='NONE') THEN
  GFTH2 = ANY(PFTH2(:,:,:)/=0.)
  GFR2  = ANY(PFR2 (:,:,:)/=0.) .AND. GUSERV
  GFTHR = ANY(PFTHR(:,:,:)/=0.) .AND. GUSERV
  GFWTH = ANY(PFWTH(:,:,:)/=0.)
  GFWR  = ANY(PFWR (:,:,:)/=0.) .AND. GUSERV
END IF
!----------------------------------------------------------------------------
!
!*       2.   SOURCES OF CONSERVATIVE POTENTIAL TEMPERATURE AND 
!                                                  PARTIAL THERMAL PRODUCTION 
!             ---------------------------------------------------------------
!
!*       2.1  Splitted value for cons. potential temperature at t+deltat
!
! Compute the turbulent flux F and F' at time t-dt.
!
#ifndef MNH_OPENACC
ZF      (:,:,:) = -XCSHF*PPHI3(:,:,:)*ZKEFF(:,:,:)*DZM(PTHLM)/PDZZ(:,:,:)
ZDFDDTDZ(:,:,:) = -XCSHF*ZKEFF(:,:,:)*D_PHI3DTDZ_O_DDTDZ(PPHI3,PREDTH1,PREDR1,PRED2TH3,PRED2THR3,HTURBDIM,GUSERV)
#else
CALL DZM_DEVICE(PTHLM, ZTMP1_DEVICE )
!$acc kernels
!$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZF      (JI,JJ,JK) = -XCSHF*PPHI3(JI,JJ,JK)*ZKEFF(JI,JJ,JK)*ZTMP1_DEVICE(JI,JJ,JK)/PDZZ(JI,JJ,JK)
!$mnh_end_do()    
!$acc end kernels
!
CALL D_PHI3DTDZ_O_DDTDZ(PPHI3,PREDTH1,PREDR1,PRED2TH3,PRED2THR3,HTURBDIM,GUSERV,ZTMP2_DEVICE)
!$acc kernels
!$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZDFDDTDZ(JI,JJ,JK) = -XCSHF*ZKEFF(JI,JJ,JK)*ZTMP2_DEVICE(JI,JJ,JK)
!$mnh_end_do()   
!$acc end kernels
#endif
!
IF (LHGRAD) THEN
 ! Compute the Leonard terms for thl
 ZDELTAX= XXHAT(3) - XXHAT(2)
 ZF_NEW (:,:,:)= XCOEFHGRADTHL*ZDELTAX*ZDELTAX/12.0*(      &
                 MXF(GX_W_UW(PWM(:,:,:), XDXX, XDZZ, XDZX))&
                *MZM(GX_M_M(PTHLM(:,:,:),XDXX,XDZZ,XDZX))  &
              +  MYF(GY_W_VW(PWM(:,:,:), XDYY,XDZZ,XDZY))  &
                *MZM(GY_M_M(PTHLM(:,:,:),XDYY,XDZZ,XDZY)) )
END IF
!
! Effect of 3rd order terms in temperature flux (at flux point)
!
! d(w'2th')/dz
#ifndef MNH_OPENACC
IF (GFWTH) THEN
  Z3RDMOMENT(:,:,:)= M3_WTH_W2TH(PREDTH1,PREDR1,PD,ZKEFF,PTKEM)
!
  ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * PFWTH(:,:,:)
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_WTH_W2TH_O_DDTDZ(PREDTH1,PREDR1,&
   & PD,PBLL_O_E,PETHETA,ZKEFF,PTKEM) * PFWTH(:,:,:)
END IF
#else
IF (GFWTH) THEN
  CALL M3_WTH_W2TH( PREDTH1, PREDR1, PD, ZKEFF, PTKEM, Z3RDMOMENT )
!
!$acc kernels
  ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * PFWTH(:,:,:)
!$acc end kernels
  CALL D_M3_WTH_W2TH_O_DDTDZ( PREDTH1, PREDR1, PD,PBLL_O_E, PETHETA, ZKEFF, PTKEM, ZTMP1_DEVICE )
!$acc kernels
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFWTH(:,:,:)
!$acc end kernels
END IF
#endif
!
! d(w'th'2)/dz
#ifndef MNH_OPENACC
IF (GFTH2) THEN
  Z3RDMOMENT(:,:,:)= M3_WTH_WTH2(PREDTH1,PREDR1,PD,PBLL_O_E,PETHETA)
!
  ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * MZM(PFTH2)
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_WTH_WTH2_O_DDTDZ(Z3RDMOMENT,PREDTH1,PREDR1,&
    & PD,PBLL_O_E,PETHETA) * MZM(PFTH2)
END IF
#else
IF (GFTH2) THEN
  CALL M3_WTH_WTH2(PREDTH1,PREDR1,PD,PBLL_O_E,PETHETA,Z3RDMOMENT)
!
  CALL MZM_DEVICE(PFTH2,ZTMP1_DEVICE)
!$acc kernels
  ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * ZTMP1_DEVICE(:,:,:)
!$acc end kernels
  CALL D_M3_WTH_WTH2_O_DDTDZ(Z3RDMOMENT,PREDTH1,PREDR1,&
    & PD,PBLL_O_E,PETHETA,ZTMP1_DEVICE)
  CALL MZM_DEVICE(PFTH2,ZTMP2_DEVICE)
!$acc kernels
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
!$acc end kernels
END IF
#endif
!
! d(w'2r')/dz
#ifndef MNH_OPENACC
IF (GFWR) THEN
  ZF(:,:,:)       = ZF(:,:,:)       + M3_WTH_W2R(PD,ZKEFF,&
    & PTKEM,PBLL_O_E,PEMOIST,PDTH_DZ) * PFWR(:,:,:)
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_WTH_W2R_O_DDTDZ(PREDTH1,PREDR1,&
    & PD,ZKEFF,PTKEM,PBLL_O_E,PEMOIST) * PFWR(:,:,:)
END IF
#else
IF (GFWR) THEN
  CALL M3_WTH_W2R( PD, ZKEFF, PTKEM, PBLL_O_E, PEMOIST, PDTH_DZ, ZTMP1_DEVICE )
!$acc kernels
  ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFWR(:,:,:)
!$acc end kernels
  CALL D_M3_WTH_W2R_O_DDTDZ( PREDTH1, PREDR1, PD, ZKEFF, PTKEM, PBLL_O_E, PEMOIST, ZTMP2_DEVICE )
!$acc kernels
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * PFWR(:,:,:)
!$acc end kernels
END IF
#endif
!
! d(w'r'2)/dz
#ifndef MNH_OPENACC
IF (GFR2) THEN
  ZF(:,:,:)       = ZF(:,:,:)       + M3_WTH_WR2(PD,ZKEFF,PTKEM,&
    & PSQRT_TKE,PBLL_O_E,PBETA,PLEPS,PEMOIST,PDTH_DZ) * MZM(PFR2)
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_WTH_WR2_O_DDTDZ(PREDTH1,PREDR1,PD,&
    & ZKEFF,PTKEM,PSQRT_TKE,PBLL_O_E,PBETA,PLEPS,PEMOIST) * MZM(PFR2)
END IF
#else
IF (GFR2) THEN
  CALL M3_WTH_WR2( PD, ZKEFF, PTKEM, PSQRT_TKE, PBLL_O_E, PBETA, PLEPS, PEMOIST, PDTH_DZ, ZTMP1_DEVICE )
  CALL MZM_DEVICE(PFR2,ZTMP2_DEVICE)
!$acc kernels
  ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
!$acc end kernels
  CALL D_M3_WTH_WR2_O_DDTDZ( PREDTH1, PREDR1, PD, ZKEFF, PTKEM, PSQRT_TKE, PBLL_O_E, PBETA, PLEPS, PEMOIST, ZTMP1_DEVICE )
  CALL MZM_DEVICE(PFR2,ZTMP2_DEVICE)
!$acc kernels
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
!$acc end kernels
END IF
#endif
!
! d(w'th'r')/dz
#ifndef MNH_OPENACC
IF (GFTHR) THEN
  Z3RDMOMENT(:,:,:)= M3_WTH_WTHR(PREDR1,PD,ZKEFF,PTKEM,PSQRT_TKE,PBETA,PLEPS,PEMOIST)
!
  ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * MZM(PFTHR)
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_WTH_WTHR_O_DDTDZ(Z3RDMOMENT,PREDTH1,&
    & PREDR1,PD,PBLL_O_E,PETHETA) * MZM(PFTHR)
END IF
#else
IF (GFTHR) THEN
  CALL M3_WTH_WTHR( PREDR1, PD, ZKEFF, PTKEM, PSQRT_TKE, PBETA, PLEPS, PEMOIST, Z3RDMOMENT )
!
  CALL MZM_DEVICE(PFTHR,ZTMP1_DEVICE)
!$acc kernels
  ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * ZTMP1_DEVICE(:,:,:)
!$acc end kernels
  CALL D_M3_WTH_WTHR_O_DDTDZ(Z3RDMOMENT,PREDTH1,PREDR1,PD,PBLL_O_E,PETHETA,ZTMP1_DEVICE)
  CALL MZM_DEVICE(PFTHR,ZTMP2_DEVICE)
!$acc kernels
  ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
!$acc end kernels
END IF
#endif
! compute interface flux
IF (LCOUPLES) THEN   ! Autocoupling O-A LES
  IF (GOCEAN) THEN    ! ocean model in coupled case 
!$acc kernels
    ZF(:,:,IKE) =  (XSSTFL_C(:,:,1)+XSSRFL_C(:,:,1)) &
                  *0.5* ( 1. + PRHODJ(:,:,KKU)/PRHODJ(:,:,IKE) )
!$acc end kernels
  ELSE                ! atmosph model in coupled case
!$acc kernels
    ZF(:,:,IKB) =  XSSTFL_C(:,:,1) &
                  *0.5* ( 1. + PRHODJ(:,:,KKA)/PRHODJ(:,:,IKB) )
!$acc end kernels
  ENDIF 
!
ELSE  ! No coupling O and A cases
  ! atmosp bottom
  !*In 3D, a part of the flux goes vertically,
  ! and another goes horizontally (in presence of slopes)
  !*In 1D, part of energy released in horizontal flux is taken into account in the vertical part
  IF (HTURBDIM=='3DIM') THEN
!$acc kernels
     !dir$ concurrent
     ZF(:,:,IKB) = ( PIMPL*PSFTHP(:,:) + PEXPL*PSFTHM(:,:) )   &
                       * PDIRCOSZW(:,:)                       &
                       * 0.5 * (1. + PRHODJ(:,:,KKA) / PRHODJ(:,:,IKB))
!$acc end kernels
  ELSE
!$acc kernels
     !dir$ concurrent
     ZF(:,:,IKB) = ( PIMPL*PSFTHP(:,:) + PEXPL*PSFTHM(:,:) )   &
                       / PDIRCOSZW(:,:)                       &
                       * 0.5 * (1. + PRHODJ(:,:,KKA) / PRHODJ(:,:,IKB))
!$acc end kernels
  END IF
!
  IF (GOCEAN) THEN
!$acc kernels present_cr(zf)
    ZF(:,:,IKE) = XSSTFL(:,:) *0.5*(1. + PRHODJ(:,:,KKU) / PRHODJ(:,:,IKE))
!$acc end kernels
  ELSE !end ocean case (in nocoupled case)
    ! atmos top
!$acc kernels
    ZF(:,:,IKE)=0.
!$acc end kernels
  END IF
END IF !end no coupled cases
!
! Compute the split conservative potential temperature at t+deltat
CALL TRIDIAG_THERMO(KKA,KKU,KKL,PTHLM,ZF,ZDFDDTDZ,PTSTEP,PIMPL,PDZZ,&
                    PRHODJ,PTHLP)
!
! Compute the equivalent tendency for the conservative potential temperature
!
!$acc kernels
!dir$ concurrent
ZRWTHL(:,:,:)= PRHODJ(:,:,:)*(PTHLP(:,:,:)-PTHLM(:,:,:))/PTSTEP
!$acc end kernels
! replace the flux by the Leonard terms above ZALT and ZCLD_THOLD
IF (LHGRAD) THEN
 DO JK=1,KKU
  ZALT(:,:,JK) = PZZ(:,:,JK)-XZS(:,:)
 END DO
 WHERE ( (ZCLD_THOLD(:,:,:) >= XCLDTHOLD) .AND. ( ZALT(:,:,:) >= XALTHGRAD) )
  ZRWTHL(:,:,:) = -GZ_W_M(MZM(PRHODJ(:,:,:))*ZF_NEW(:,:,:),XDZZ)
 END WHERE
END IF
!
!$acc kernels
PRTHLS(:,:,:)= PRTHLS(:,:,:)  + ZRWTHL(:,:,:)
!$acc end kernels
!
!*       2.2  Partial Thermal Production
!
!  Conservative potential temperature flux : 
!
#ifndef MNH_OPENACC
ZFLXZ(:,:,:)   = ZF(:,:,:)                                                &
               + PIMPL * ZDFDDTDZ(:,:,:) * DZM(PTHLP - PTHLM(:,:,:)) / PDZZ(:,:,:)
#else
!$acc kernels
!dir$ concurrent
ZTMP1_DEVICE(:,:,:) = PTHLP(:,:,:) - PTHLM(:,:,:)
!$acc end kernels
CALL DZM_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
!$acc kernels
!$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZFLXZ(JI,JJ,JK)   =  ZF(JI,JJ,JK) + PIMPL * ZDFDDTDZ(JI,JJ,JK) * ZTMP2_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK)
!$mnh_end_do()
!$acc end kernels
#endif
! replace the flux by the Leonard terms
IF (LHGRAD) THEN
 WHERE ( (ZCLD_THOLD(:,:,:) >= XCLDTHOLD) .AND. ( ZALT(:,:,:) >= XALTHGRAD) )
  ZFLXZ(:,:,:) = ZF_NEW(:,:,:)
 END WHERE
END IF
!
!$acc kernels
ZFLXZ(:,:,KKA) = ZFLXZ(:,:,IKB)
IF (GOCEAN) THEN
  ZFLXZ(:,:,KKU) = ZFLXZ(:,:,IKE)
END IF
!  
PWTH(:,:,:IKTB) = XUNDEF
PWTH(:,:,IKTE:) = XUNDEF
!
DO JK=IKTB+1,IKTE-1
  PWTH(:,:,JK)=0.5*(ZFLXZ(:,:,JK)+ZFLXZ(:,:,JK+KKL))
END DO
!
PWTH(:,:,IKB)=0.5*(ZFLXZ(:,:,IKB)+ZFLXZ(:,:,IKB+KKL)) 
!
IF (GOCEAN) THEN
  PWTH(:,:,IKE)=0.5*(ZFLXZ(:,:,IKE)+ZFLXZ(:,:,IKE+KKL))
  PWTH(:,:,KKA)=0. 
  PWTH(:,:,KKU)=ZFLXZ(:,:,KKU)
ELSE
  PWTH(:,:,IKE)=PWTH(:,:,IKE-KKL)
  PWTH(:,:,KKA)=0.5*(ZFLXZ(:,:,KKA)+ZFLXZ(:,:,KKA+KKL))
END IF
!$acc end kernels
!
IF ( OTURB_FLX .AND. tpfile%lopened ) THEN
!$acc update self(ZFLXZ)
  ! stores the conservative potential temperature vertical flux
  TZFIELD%CMNHNAME   = 'THW_FLX'
  TZFIELD%CSTDNAME   = ''
  TZFIELD%CLONGNAME  = 'THW_FLX'
  TZFIELD%CUNITS     = 'K m s-1'
  TZFIELD%CDIR       = 'XY'
  TZFIELD%CCOMMENT   = 'Conservative potential temperature vertical flux'
  TZFIELD%NGRID      = 4
  TZFIELD%NTYPE      = TYPEREAL
  TZFIELD%NDIMS      = 3
  TZFIELD%LTIMEDEP   = .TRUE.
  CALL IO_Field_write(TPFILE,TZFIELD,ZFLXZ)
END IF
!
! Contribution of the conservative temperature flux to the buoyancy flux
IF (GOCEAN) THEN
#ifndef MNH_OPENACC
  PTP(:,:,:)= XG*XALPHAOC * MZF(ZFLXZ )
#else
  CALL MZF_DEVICE( ZFLXZ, ZTMP1_DEVICE )
  PTP(:,:,:) = XG * XALPHAOC * ZTMP1_DEVICE
#endif
ELSE
#ifndef MNH_OPENACC
  IF (KRR /= 0) THEN
    PTP(:,:,:)  =  PBETA(:,:,:) * MZF( MZM(PETHETA) * ZFLXZ(:,:,:) )
    PTP(:,:,IKB)=  PBETA(:,:,IKB) * PETHETA(:,:,IKB) *   &
                   0.5 * ( ZFLXZ(:,:,IKB) + ZFLXZ(:,:,IKB+KKL) )
  ELSE
    PTP(:,:,:)=  PBETA(:,:,:) * MZF( ZFLXZ )
  END IF
#else
  IF (KRR /= 0) THEN
    CALL MZM_DEVICE(PETHETA,ZTMP1_DEVICE)
!$acc kernels
    !$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
      ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) * ZFLXZ(JI,JJ,JK)
    !$mnh_end_do()
!$acc end kernels
    CALL MZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
!$acc kernels
    PTP(:,:,:)  =  PBETA(:,:,:) * ZTMP3_DEVICE(:,:,:)
    PTP(:,:,IKB)=  PBETA(:,:,IKB) * PETHETA(:,:,IKB) *   &
                   0.5 * ( ZFLXZ(:,:,IKB) + ZFLXZ(:,:,IKB+KKL) )
!$acc end kernels 
  ELSE
    CALL MZF_DEVICE( ZFLXZ, ZTMP1_DEVICE )
!$acc kernels
    PTP(:,:,:)=  PBETA(:,:,:) * ZTMP1_DEVICE(:,:,:)
!$acc end kernels
  END IF
#endif
END IF
!
! Buoyancy flux at flux points
! 
#ifndef MNH_OPENACC
PWTHV(:,:,:) = MZM(PETHETA) * ZFLXZ(:,:,:)
#else
CALL MZM_DEVICE(PETHETA,ZTMP1_DEVICE)
!$acc kernels
PWTHV(:,:,:) = ZTMP1_DEVICE(:,:,:) * ZFLXZ(:,:,:)
#endif
PWTHV(:,:,IKB) = PETHETA(:,:,IKB) * ZFLXZ(:,:,IKB)
!$acc end kernels
!
IF (GOCEAN) THEN
  ! temperature contribution to Buy flux     
!$acc kernels
  PWTHV(:,:,IKE) = PETHETA(:,:,IKE) * ZFLXZ(:,:,IKE)
!$acc end kernels
END IF
!*       2.3  Partial vertical divergence of the < Rc w > flux
!
#ifndef MNH_OPENACC
IF ( KRRL >= 1 ) THEN
  IF ( KRRI >= 1 ) THEN
    PRRS(:,:,:,2) = PRRS(:,:,:,2) -                                                            &
                    PRHODJ(:,:,:)*PATHETA(:,:,:)*2.*PSRCM(:,:,:)*DZF(ZFLXZ(:,:,:)/PDZZ(:,:,:)) &
                    *(1.0-PFRAC_ICE(:,:,:))
    PRRS(:,:,:,4) = PRRS(:,:,:,4) -                                                            &
                    PRHODJ(:,:,:)*PATHETA(:,:,:)*2.*PSRCM(:,:,:)*DZF(ZFLXZ(:,:,:)/PDZZ(:,:,:)) &
                    *PFRAC_ICE(:,:,:)
  ELSE
    PRRS(:,:,:,2) = PRRS(:,:,:,2) -                                                            &
                    PRHODJ(:,:,:)*PATHETA(:,:,:)*2.*PSRCM(:,:,:)*DZF(ZFLXZ(:,:,:)/PDZZ(:,:,:))
  END IF
END IF
#else
IF ( KRRL >= 1 ) THEN
  IF ( KRRI >= 1 ) THEN
!Remark: to work around a problem with NVHPC 23.7 (not present in previous NVHPC versions),
! the computation of ZTMP1_DEVICE and ZTMP2_DEVICE has been split in 2 different kernels
!$acc kernels
    !dir$ concurrent
    ZTMP1_DEVICE(:,:,:) = PRHODJ(:,:,:) * PATHETA(:,:,:) * 2. * PSRCM(:,:,:)
!$acc end kernels
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP2_DEVICE)
#else
!$acc kernels
#endif
    !dir$ concurrent
    ZTMP2_DEVICE(:,:,:) = ZFLXZ(:,:,:) / PDZZ(:,:,:)
!$acc end kernels
    CALL DZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
!$acc kernels
    PRRS(:,:,:,2) = PRRS(:,:,:,2) - ZTMP1_DEVICE(:,:,:) * ZTMP3_DEVICE * (1.0-PFRAC_ICE(:,:,:))
    PRRS(:,:,:,4) = PRRS(:,:,:,4) - ZTMP1_DEVICE(:,:,:) * ZTMP3_DEVICE * PFRAC_ICE(:,:,:)
!$acc end kernels
  ELSE
!Remark: to work around a problem with NVHPC 23.7 (not present in previous NVHPC versions),
! the computation of ZTMP1_DEVICE and ZTMP2_DEVICE has been split in 2 different kernels
!$acc kernels
    !dir$ concurrent
    ZTMP1_DEVICE(:,:,:) = PRHODJ(:,:,:) * PATHETA(:,:,:) * 2. * PSRCM(:,:,:)
!$acc end kernels
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP2_DEVICE)
#else
!$acc kernels
#endif
    !dir$ concurrent
    ZTMP2_DEVICE(:,:,:) = ZFLXZ(:,:,:) / PDZZ(:,:,:)
!$acc end kernels
    CALL DZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
!$acc kernels
    PRRS(:,:,:,2) = PRRS(:,:,:,2) - ZTMP1_DEVICE(:,:,:) * ZTMP3_DEVICE(:,:,:)
!$acc end kernels
  END IF
END IF
#endif
!
!*       2.4  Storage in LES configuration
! 
IF (LLES_CALL) THEN
  CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
  CALL LES_MEAN_SUBGRID( MZF(ZFLXZ), X_LES_SUBGRID_WThl )
  CALL LES_MEAN_SUBGRID( MZF(PWM*ZFLXZ), X_LES_RES_W_SBG_WThl )
  CALL LES_MEAN_SUBGRID( GZ_W_M(PWM,PDZZ)*MZF(ZFLXZ),&
      & X_LES_RES_ddxa_W_SBG_UaThl )
  CALL LES_MEAN_SUBGRID( MZF(PDTH_DZ*ZFLXZ), X_LES_RES_ddxa_Thl_SBG_UaThl )
  CALL LES_MEAN_SUBGRID( -XCTP*PSQRT_TKE/PLM*MZF(ZFLXZ), X_LES_SUBGRID_ThlPz )
  CALL LES_MEAN_SUBGRID( MZF(MZM(PETHETA)*ZFLXZ), X_LES_SUBGRID_WThv )
  IF (KRR>=1) THEN
    CALL LES_MEAN_SUBGRID( MZF(PDR_DZ*ZFLXZ), X_LES_RES_ddxa_Rt_SBG_UaThl )
  END IF
  !* diagnostic of mixing coefficient for heat
  ZA = DZM(PTHLP)
  WHERE (ZA==0.) ZA=1.E-6
  ZA = - ZFLXZ / ZA * PDZZ
  ZA(:,:,IKB) = XCSHF*PPHI3(:,:,IKB)*ZKEFF(:,:,IKB)
  ZA = MZF( ZA )
  ZA = MIN(MAX(ZA,-1000.),1000.)
  CALL LES_MEAN_SUBGRID( ZA, X_LES_SUBGRID_Kh   ) 
#else
!$acc data copy(X_LES_SUBGRID_WThl,X_LES_RES_W_SBG_WThl,X_LES_RES_ddxa_W_SBG_UaThl,X_LES_RES_ddxa_Thl_SBG_UaThl,&
!$acc &         X_LES_SUBGRID_ThlPz,X_LES_SUBGRID_WThv,X_LES_RES_ddxa_Rt_SBG_UaThl,X_LES_SUBGRID_Kh)
  CALL MZF_DEVICE( ZFLXZ, ZTMP1_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_SUBGRID_WThl )
!$acc kernels
  ZTMP1_DEVICE(:,:,:) = PWM(:,:,:)*ZFLXZ(:,:,:)
!$acc end kernels
  CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_RES_W_SBG_WThl )
  CALL GZ_W_M_DEVICE(PWM,PDZZ,ZTMP1_DEVICE )
  CALL MZF_DEVICE( ZFLXZ, ZTMP2_DEVICE )
!$acc kernels
  ZTMP3_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
!$acc end kernels
  CALL LES_MEAN_SUBGRID( ZTMP3_DEVICE, X_LES_RES_ddxa_W_SBG_UaThl )
!$acc kernels
  ZTMP1_DEVICE(:,:,:) = PDTH_DZ(:,:,:)*ZFLXZ(:,:,:)
!$acc end kernels
  CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_RES_ddxa_Thl_SBG_UaThl )
  CALL MZF_DEVICE( ZFLXZ, ZTMP1_DEVICE )
!$acc kernels
  ZTMP2_DEVICE(:,:,:) = -XCTP*PSQRT_TKE(:,:,:)/PLM(:,:,:)*ZTMP1_DEVICE(:,:,:)
!$acc end kernels
  CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_SUBGRID_ThlPz ) 
  CALL MZM_DEVICE(PETHETA,ZTMP1_DEVICE)
!$acc kernels
  ZTMP2_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZFLXZ(:,:,:)
!$acc end kernels
  CALL MZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP3_DEVICE, X_LES_SUBGRID_WThv ) 
  IF (KRR>=1) THEN
!$acc kernels
    ZTMP1_DEVICE(:,:,:) = PDR_DZ(:,:,:)*ZFLXZ(:,:,:)
!$acc end kernels
    CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
    CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_RES_ddxa_Rt_SBG_UaThl )
  END IF
  !* diagnostic of mixing coefficient for heat
  CALL DZM_DEVICE( PTHLP, ZA )
  !$acc kernels
  !$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     IF  (ZA(JI,JJ,JK)==0.) THEN
        ZA(JI,JJ,JK)=1.E-6
     END IF
  !$mnh_end_do()
  ZA(:,:,:) = - ZFLXZ(:,:,:) / ZA(:,:,:) * PDZZ(:,:,:)
  ZA(:,:,IKB) = XCSHF*PPHI3(:,:,IKB)*ZKEFF(:,:,IKB)
  !Copy ZA into ZTMP1_DEVICE to prevent aliasing in the following call to MZF_DEVICE
  ZTMP1_DEVICE(:,:,:) = ZA(:,:,:)
!$acc end kernels
  CALL MZF_DEVICE( ZTMP1_DEVICE, ZA )
!$acc kernels
  ZA(:,:,:) = MIN(MAX(ZA(:,:,:),-1000.),1000.)
!$acc end kernels
  CALL LES_MEAN_SUBGRID( ZA, X_LES_SUBGRID_Kh   ) 
!$acc end data
#endif
  !
  CALL SECOND_MNH(ZTIME2)
  XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
END IF
!
!*       2.5  New boundary layer depth for TOMs
! 
IF (HTOM=='TM06') CALL TM06_H(IKB,IKTB,IKTE,PTSTEP,PZZ,ZFLXZ,PBL_DEPTH)
!
!----------------------------------------------------------------------------
!
!
!*       3.   SOURCES OF CONSERVATIVE AND CLOUD MIXING RATIO AND 
!                                        COMPLETE THERMAL PRODUCTION 
!             ------------------------------------------------------
!
!*       3.1  Splitted value for cons. mixing ratio at t+deltat
!
!
IF (KRR /= 0) THEN
  ! Compute the turbulent flux F and F' at time t-dt.
  !
#ifndef MNH_OPENACC
  ZF      (:,:,:) = -XCSHF*PPSI3(:,:,:)*ZKEFF(:,:,:)*DZM(PRM(:,:,:,1))/PDZZ(:,:,:)
  ZDFDDRDZ(:,:,:) = -XCSHF*ZKEFF(:,:,:)*D_PSI3DRDZ_O_DDRDZ(PPSI3,PREDR1,PREDTH1,PRED2R3,PRED2THR3,HTURBDIM,GUSERV)
#else
  CALL DZM_DEVICE( PRM(:,:,:,1), ZTMP1_DEVICE )
  !$acc kernels
  !$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZF      (JI,JJ,JK) = -XCSHF*PPSI3(JI,JJ,JK)*ZKEFF(JI,JJ,JK)*ZTMP1_DEVICE(JI,JJ,JK)/PDZZ(JI,JJ,JK)
  !$mnh_end_do()
  !$acc end kernels
  CALL D_PSI3DRDZ_O_DDRDZ(PPSI3,PREDR1,PREDTH1,PRED2R3,PRED2THR3,HTURBDIM,GUSERV,ZTMP1_DEVICE)
!CALL D_PHI3DRDZ_O_DDRDZ_DEVICE(PPSI3,PREDR1,PREDTH1,PRED2R3,PRED2THR3,HTURBDIM,GUSERV,ZTMP1_DEVICE)
  !$acc kernels
  !$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZDFDDRDZ(JI,JJ,JK) = -XCSHF*ZKEFF(JI,JJ,JK)*ZTMP1_DEVICE(JI,JJ,JK)
  !$mnh_end_do()
  !$acc end kernels
#endif
  !
  ! Compute Leonard Terms for Cloud mixing ratio
  IF (LHGRAD) THEN
    ZDELTAX= XXHAT(3) - XXHAT(2)
    ZF_NEW (:,:,:)= XCOEFHGRADRM*ZDELTAX*ZDELTAX/12.0*(        &
                MXF(GX_W_UW(PWM(:,:,:), XDXX, XDZZ, XDZX))       &
                *MZM(GX_M_M(PRM(:,:,:,1),XDXX,XDZZ,XDZX)) &
                +MYF(GY_W_VW(PWM(:,:,:), XDYY,XDZZ,XDZY))        &
                *MZM(GY_M_M(PRM(:,:,:,1),XDYY,XDZZ,XDZY)) )
   END IF
  !
  ! Effect of 3rd order terms in temperature flux (at flux point)
  !
  ! d(w'2r')/dz
#ifndef MNH_OPENACC
  IF (GFWR) THEN
    Z3RDMOMENT(:,:,:)= M3_WR_W2R(PREDR1,PREDTH1,PD,ZKEFF,PTKEM)
  !
    ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * PFWR(:,:,:)
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_WR_W2R_O_DDRDZ(PREDR1,PREDTH1,PD,&
     & PBLL_O_E,PEMOIST,ZKEFF,PTKEM) * PFWR(:,:,:)
  END IF
#else
  IF (GFWR) THEN
    CALL M3_WR_W2R(PREDR1,PREDTH1,PD,ZKEFF,PTKEM,Z3RDMOMENT)
  !
  !$acc kernels
    ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * PFWR(:,:,:)
  !$acc end kernels
    CALL D_M3_WR_W2R_O_DDRDZ( PREDR1, PREDTH1, PD, PBLL_O_E, PEMOIST, ZKEFF, PTKEM, ZTMP1_DEVICE )
  !$acc kernels
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFWR(:,:,:)
  !$acc end kernels
  END IF
#endif
  !
  ! d(w'r'2)/dz
#ifndef MNH_OPENACC
  IF (GFR2) THEN
    Z3RDMOMENT(:,:,:)= M3_WR_WR2(PREDR1,PREDTH1,PD,PBLL_O_E,PEMOIST)
  !
    ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * MZM(PFR2)
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_WR_WR2_O_DDRDZ(Z3RDMOMENT,PREDR1,&
     & PREDTH1,PD,PBLL_O_E,PEMOIST) * MZM(PFR2)
  END IF
#else
  IF (GFR2) THEN
    CALL M3_WR_WR2(PREDR1,PREDTH1,PD,PBLL_O_E,PEMOIST,Z3RDMOMENT)
  !
    CALL MZM_DEVICE(PFR2,ZTMP1_DEVICE)
  !$acc kernels
    ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * ZTMP1_DEVICE(:,:,:)
  !$acc end kernels
    CALL D_M3_WR_WR2_O_DDRDZ(Z3RDMOMENT,PREDR1,&
     & PREDTH1,PD,PBLL_O_E,PEMOIST,ZTMP2_DEVICE)
  !$acc kernels
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
  !$acc end kernels
  END IF
#endif
  !
  ! d(w'2th')/dz
#ifndef MNH_OPENACC
  IF (GFWTH) THEN
    ZF(:,:,:)       = ZF(:,:,:)       + M3_WR_W2TH(PD,ZKEFF,&
     & PTKEM,PBLL_O_E,PETHETA,PDR_DZ) * PFWTH(:,:,:)
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_WR_W2TH_O_DDRDZ(PREDR1,PREDTH1,&
     & PD,ZKEFF,PTKEM,PBLL_O_E,PETHETA) * PFWTH(:,:,:)
  END IF
#else
  IF (GFWTH) THEN
    CALL M3_WR_W2TH( PD,ZKEFF,&
     & PTKEM,PBLL_O_E,PETHETA,PDR_DZ,ZTMP1_DEVICE)
  !$acc kernels
    ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFWTH(:,:,:)
  !$acc end kernels
    CALL D_M3_WR_W2TH_O_DDRDZ( PREDR1, PREDTH1, PD, ZKEFF, PTKEM, PBLL_O_E, PETHETA, ZTMP1_DEVICE )
  !$acc kernels
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFWTH(:,:,:)
  !$acc end kernels
  END IF
#endif
  !
  ! d(w'th'2)/dz
#ifndef MNH_OPENACC
  IF (GFTH2) THEN
    ZF(:,:,:)       = ZF(:,:,:)       + M3_WR_WTH2(PD,ZKEFF,PTKEM,&
    & PSQRT_TKE,PBLL_O_E,PBETA,PLEPS,PETHETA,PDR_DZ) * MZM(PFTH2)
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_WR_WTH2_O_DDRDZ(PREDR1,PREDTH1,PD,&
     &ZKEFF,PTKEM,PSQRT_TKE,PBLL_O_E,PBETA,PLEPS,PETHETA) * MZM(PFTH2)
  END IF
#else
  IF (GFTH2) THEN
    CALL M3_WR_WTH2( PD, ZKEFF, PTKEM, PSQRT_TKE, PBLL_O_E, PBETA, PLEPS, PETHETA, PDR_DZ, ZTMP1_DEVICE )
    CALL MZM_DEVICE(PFTH2,ZTMP2_DEVICE)
  !$acc kernels
    ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
  !$acc end kernels
    CALL D_M3_WR_WTH2_O_DDRDZ( PREDR1, PREDTH1, PD, ZKEFF, PTKEM, PSQRT_TKE, PBLL_O_E, PBETA, PLEPS, PETHETA, ZTMP1_DEVICE )
  !$acc kernels
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
  !$acc end kernels
  END IF
#endif
  !
  ! d(w'th'r')/dz
#ifndef MNH_OPENACC
  IF (GFTHR) THEN
    Z3RDMOMENT(:,:,:)= M3_WR_WTHR(PREDTH1,PD,ZKEFF,PTKEM,PSQRT_TKE,PBETA,&
     & PLEPS,PETHETA)
  !
    ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * MZM(PFTHR)
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_WR_WTHR_O_DDRDZ(Z3RDMOMENT,PREDR1, &
     & PREDTH1,PD,PBLL_O_E,PEMOIST) * MZM(PFTHR)
  END IF
#else
  IF (GFTHR) THEN
    CALL M3_WR_WTHR( PREDTH1, PD, ZKEFF, PTKEM, PSQRT_TKE, PBETA, PLEPS, PETHETA, Z3RDMOMENT )
  !
    CALL MZM_DEVICE(PFTHR,ZTMP1_DEVICE)
  !$acc kernels
    ZF(:,:,:)       = ZF(:,:,:)       + Z3RDMOMENT(:,:,:) * ZTMP1_DEVICE(:,:,:)
  !$acc end kernels
    CALL D_M3_WR_WTHR_O_DDRDZ( Z3RDMOMENT,PREDR1, &
     & PREDTH1,PD,PBLL_O_E,PEMOIST,ZTMP2_DEVICE)
  !$acc kernels
    ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
  !$acc end kernels
  END IF
#endif
  !
  ! compute interface flux
  IF (LCOUPLES) THEN   ! coupling NH O-A
    IF (GOCEAN) THEN    ! ocean model in coupled case
      ! evap effect on salinity to be added later !!!
!$acc kernels
      ZF(:,:,IKE) =  0.
!$acc end kernels
    ELSE                ! atmosph model in coupled case
!$acc kernels
      ZF(:,:,IKB) =  0.
!$acc end kernels
      ! AJOUTER FLUX EVAP SUR MODELE ATMOS
    ENDIF
  !
  ELSE  ! No coupling NH OA case
    ! atmosp bottom
    !* in 3DIM case, a part of the flux goes vertically, and another goes horizontally
    ! (in presence of slopes)
    !* in 1DIM case, the part of energy released in horizontal flux
    ! is taken into account in the vertical part
    !
    IF (HTURBDIM=='3DIM') THEN
!$acc kernels present_cr(zf)
      ZF(:,:,IKB) = ( PIMPL*PSFRP(:,:) + PEXPL*PSFRM(:,:) )       &
                           * PDIRCOSZW(:,:)                       &
                         * 0.5 * (1. + PRHODJ(:,:,KKA) / PRHODJ(:,:,IKB))
!$acc end kernels
    ELSE
!$acc kernels present_cr(zf)
      ZF(:,:,IKB) = ( PIMPL*PSFRP(:,:) + PEXPL*PSFRM(:,:) )     &
                         / PDIRCOSZW(:,:)                       &
                         * 0.5 * (1. + PRHODJ(:,:,KKA) / PRHODJ(:,:,IKB))
!$acc end kernels
    END IF
    !
    IF (GOCEAN) THEN
      ! General ocean case
      ! salinity/evap effect to be added later !!!!!
!$acc kernels
      ZF(:,:,IKE) = 0.
!$acc end kernels
    ELSE !end ocean case (in nocoupled case)
      ! atmos top
!$acc kernels
     ZF(:,:,IKE)=0.
!$acc end kernels
    END IF
  END IF!end no coupled cases
  ! Compute the split conservative potential temperature at t+deltat
  CALL TRIDIAG_THERMO(KKA,KKU,KKL,PRM(:,:,:,1),ZF,ZDFDDRDZ,PTSTEP,PIMPL,&
                      PDZZ,PRHODJ,PRP)
  !
  ! Compute the equivalent tendency for the conservative mixing ratio
  !
  !$acc kernels
  !dir$ concurrent
  ZRWRNP (:,:,:) = PRHODJ(:,:,:)*(PRP(:,:,:)-PRM(:,:,:,1))/PTSTEP
  !$acc end kernels
  !
  ! replace the flux by the Leonard terms above ZALT and ZCLD_THOLD
  IF (LHGRAD) THEN
   DO JK=1,KKU
    ZALT(:,:,JK) = PZZ(:,:,JK)-XZS(:,:)
   END DO
   WHERE ( (ZCLD_THOLD(:,:,:) >= XCLDTHOLD ) .AND. ( ZALT(:,:,:) >= XALTHGRAD ) )
    ZRWRNP (:,:,:) =  -GZ_W_M(MZM(PRHODJ(:,:,:))*ZF_NEW(:,:,:),XDZZ)
   END WHERE
  END IF
  !
!$acc kernels
  PRRS(:,:,:,1) = PRRS(:,:,:,1) + ZRWRNP (:,:,:)
!$acc end kernels
  !
  !*       3.2  Complete thermal production
  !
  ! cons. mixing ratio flux :
  !
#ifndef MNH_OPENACC
  ZFLXZ(:,:,:)   = ZF(:,:,:)                                                &
                 + PIMPL * ZDFDDRDZ(:,:,:) * DZM(PRP - PRM(:,:,:,1)) / PDZZ(:,:,:)
#else
!$acc kernels
  !dir$ concurrent
  ZTMP1_DEVICE(:,:,:) = PRP(:,:,:) - PRM(:,:,:,1)
!$acc end kernels
  CALL DZM_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
!$acc kernels
!$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)  
   ZFLXZ(JI,JJ,JK)   = ZF(JI,JJ,JK) + PIMPL * ZDFDDRDZ(JI,JJ,JK) *ZTMP2_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK)
!$mnh_end_do()
!$acc end kernels
#endif
  !
  ! replace the flux by the Leonard terms above ZALT and ZCLD_THOLD
  IF (LHGRAD) THEN
!$acc kernels present_cr(zf_new,zflxz)
   WHERE ( (ZCLD_THOLD(:,:,:) >= ZCLDTHOLD ) .AND. ( ZALT(:,:,:) >= ZALTHGRAD ) )
    ZFLXZ(:,:,:) = ZF_NEW(:,:,:)
   END WHERE
!$acc end kernels
  END IF
  !
!$acc kernels
  ZFLXZ(:,:,KKA) = ZFLXZ(:,:,IKB) 
  !
  DO JK=IKTB+1,IKTE-1
   PWRC(:,:,JK)=0.5*(ZFLXZ(:,:,JK)+ZFLXZ(:,:,JK+KKL))
  END DO
  PWRC(:,:,IKB)=0.5*(ZFLXZ(:,:,IKB)+ZFLXZ(:,:,IKB+KKL))
  PWRC(:,:,KKA)=0.5*(ZFLXZ(:,:,KKA)+ZFLXZ(:,:,KKA+KKL))
  PWRC(:,:,IKE)=PWRC(:,:,IKE-KKL)
  PWRC(:, :, IKE + 1 : ) = XUNDEF
!$acc end kernels
  !
  !
  IF ( OTURB_FLX .AND. tpfile%lopened ) THEN
!$acc update self(ZFLXZ)
    ! stores the conservative mixing ratio vertical flux
    TZFIELD%CMNHNAME   = 'RCONSW_FLX'
    TZFIELD%CSTDNAME   = ''
    TZFIELD%CLONGNAME  = 'RCONSW_FLX'
    TZFIELD%CUNITS     = 'kg m s-1 kg-1'
    TZFIELD%CDIR       = 'XY'
    TZFIELD%CCOMMENT   = 'Conservative mixing ratio vertical flux'
    TZFIELD%NGRID      = 4
    TZFIELD%NTYPE      = TYPEREAL
    TZFIELD%NDIMS      = 3
    TZFIELD%LTIMEDEP   = .TRUE.
    CALL IO_Field_write(TPFILE,TZFIELD,ZFLXZ)
  END IF
  !
  ! Contribution of the conservative water flux to the Buoyancy flux
  IF (GOCEAN) THEN
#ifndef MNH_OPENACC
    ZA(:,:,:)=  -XG * XBETAOC * MZF(ZFLXZ )
#else
    CALL MZF_DEVICE( ZFLXZ, ZTMP1_DEVICE )
!$acc kernels
    ZA(:,:,:)=  -XG * XBETAOC * ZTMP1_DEVICE(:,:,:)
!$acc end kernels
#endif
  ELSE
#ifndef MNH_OPENACC
    ZA(:,:,:)   =  PBETA * MZF( MZM(PEMOIST) * ZFLXZ )
#else
    CALL MZM_DEVICE(PEMOIST,ZTMP1_DEVICE)
!$acc kernels
    !$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
      ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) * ZFLXZ(JI,JJ,JK)
    !$mnh_end_do()
!$acc end kernels
    CALL MZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
!$acc kernels
    !$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
      ZA(JI,JJ,JK)   =  PBETA(JI,JJ,JK) * ZTMP3_DEVICE(JI,JJ,JK)
    !$mnh_end_do()
!$acc end kernels
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZA)
#else
!$acc kernels 
#endif
#endif
    ZA(:,:,IKB) =  PBETA(:,:,IKB) * PEMOIST(:,:,IKB) *   &
                   0.5 * ( ZFLXZ (:,:,IKB) + ZFLXZ (:,:,IKB+KKL) )
    PTP(:,:,:) = PTP(:,:,:) + ZA(:,:,:)
!$acc end kernels
  END IF
  !
  ! Buoyancy flux at flux points
  ! 
#ifndef MNH_OPENACC
  PWTHV          = PWTHV          + MZM(PEMOIST) * ZFLXZ
#else
  CALL MZM_DEVICE(PEMOIST,ZTMP1_DEVICE)
  !$acc kernels
  !$mnh_do_concurrent ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     PWTHV(JI,JJ,JK)          = PWTHV(JI,JJ,JK)          + ZTMP1_DEVICE(JI,JJ,JK) * ZFLXZ(JI,JJ,JK)
  !$mnh_end_do()
  !$acc end kernels
#endif
  !$acc kernels
  PWTHV(:,:,IKB) = PWTHV(:,:,IKB) + PEMOIST(:,:,IKB) * ZFLXZ(:,:,IKB)
!$acc end kernels
  IF (GOCEAN) THEN
!$acc kernels
    PWTHV(:,:,IKE) = PWTHV(:,:,IKE) + PEMOIST(:,:,IKE)* ZFLXZ(:,:,IKE)
!$acc end kernels
  END IF   
!
!*       3.3  Complete vertical divergence of the < Rc w > flux
!
#ifndef MNH_OPENACC
  IF ( KRRL >= 1 ) THEN
    IF ( KRRI >= 1 ) THEN
      PRRS(:,:,:,2) = PRRS(:,:,:,2) -                                        &
                      PRHODJ*PAMOIST*2.*PSRCM*DZF(ZFLXZ/PDZZ )       &
                      *(1.0-PFRAC_ICE(:,:,:))
      PRRS(:,:,:,4) = PRRS(:,:,:,4) -                                        &
                      PRHODJ*PAMOIST*2.*PSRCM*DZF(ZFLXZ/PDZZ )       &
                      *PFRAC_ICE(:,:,:)
    ELSE
      PRRS(:,:,:,2) = PRRS(:,:,:,2) -                                        &
                      PRHODJ*PAMOIST*2.*PSRCM*DZF(ZFLXZ/PDZZ )
    END IF
  END IF
#else
  IF ( KRRL >= 1 ) THEN
    IF ( KRRI >= 1 ) THEN
!Remark: to work around a problem with NVHPC 23.7 (not present in previous NVHPC versions),
! the computation of ZTMP1_DEVICE and ZTMP2_DEVICE has been split in 2 different kernels
!$acc kernels
      !dir$ concurrent
      ZTMP1_DEVICE(:,:,:) = PRHODJ(:,:,:)*PAMOIST(:,:,:)*2.*PSRCM(:,:,:)
!$acc end kernels
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP2_DEVICE)
#else
!$acc kernels
#endif
      !dir$ concurrent
      ZTMP2_DEVICE(:,:,:) = ZFLXZ(:,:,:)/PDZZ(:,:,:)
!$acc end kernels
      CALL DZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
!$acc kernels
      PRRS(:,:,:,2) = PRRS(:,:,:,2) - ZTMP1_DEVICE(:,:,:) * ZTMP3_DEVICE(:,:,:) * (1.0-PFRAC_ICE(:,:,:))
      PRRS(:,:,:,4) = PRRS(:,:,:,4) - ZTMP1_DEVICE(:,:,:) * ZTMP3_DEVICE(:,:,:) * PFRAC_ICE(:,:,:)
!$acc end kernels
    ELSE
!Remark: to work around a problem with NVHPC 23.7 (not present in previous NVHPC versions),
! the computation of ZTMP1_DEVICE and ZTMP2_DEVICE has been split in 2 different kernels
!$acc kernels
      !dir$ concurrent
      ZTMP1_DEVICE(:,:,:) = PRHODJ(:,:,:)*PAMOIST(:,:,:)*2.*PSRCM(:,:,:)
!$acc end kernels
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP2_DEVICE)
#else
!$acc kernels
#endif
      !dir$ concurrent
      ZTMP2_DEVICE(:,:,:) = ZFLXZ(:,:,:)/PDZZ(:,:,:)
!$acc end kernels
      CALL DZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
!$acc kernels
      PRRS(:,:,:,2) = PRRS(:,:,:,2) - ZTMP1_DEVICE(:,:,:) * ZTMP3_DEVICE(:,:,:)
!$acc end kernels
    END IF
  END IF
#endif
!
!*       3.4  Storage in LES configuration
! 
  IF (LLES_CALL) THEN
    CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
    CALL LES_MEAN_SUBGRID( MZF(ZFLXZ), X_LES_SUBGRID_WRt )
    CALL LES_MEAN_SUBGRID( MZF(PWM(:,:,:)*ZFLXZ(:,:,:)), X_LES_RES_W_SBG_WRt )
    CALL LES_MEAN_SUBGRID( GZ_W_M(PWM,PDZZ)*MZF(ZFLXZ),&
    & X_LES_RES_ddxa_W_SBG_UaRt )
    CALL LES_MEAN_SUBGRID( MZF(PDTH_DZ(:,:,:)*ZFLXZ(:,:,:)), X_LES_RES_ddxa_Thl_SBG_UaRt )
    CALL LES_MEAN_SUBGRID( MZF(PDR_DZ(:,:,:)*ZFLXZ(:,:,:)), X_LES_RES_ddxa_Rt_SBG_UaRt )
    CALL LES_MEAN_SUBGRID( MZF(MZM(PEMOIST)*ZFLXZ(:,:,:)), X_LES_SUBGRID_WThv , .TRUE. )
    CALL LES_MEAN_SUBGRID( -XCTP*PSQRT_TKE(:,:,:)/PLM(:,:,:)*MZF(ZFLXZ), X_LES_SUBGRID_RtPz )
#else
!$acc data copy(X_LES_SUBGRID_WRt,X_LES_RES_W_SBG_WRt,X_LES_RES_ddxa_W_SBG_UaRt,X_LES_RES_ddxa_Thl_SBG_UaRt,&
!$acc &         X_LES_RES_ddxa_Rt_SBG_UaRt,X_LES_SUBGRID_WThv,X_LES_SUBGRID_RtPz)
    CALL MZF_DEVICE( ZFLXZ, ZTMP1_DEVICE )
    CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_SUBGRID_WRt )
    !
    !$acc kernels
    ZTMP2_DEVICE(:,:,:) = PWM(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL MZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
    CALL LES_MEAN_SUBGRID( ZTMP3_DEVICE, X_LES_RES_W_SBG_WRt )
    !
    !$acc kernels
    ZTMP1_DEVICE(:,:,:) = PWM(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL GZ_W_M_DEVICE(PWM,PDZZ,ZTMP2_DEVICE)
    !$acc kernels
    ZTMP3_DEVICE(:,:,:) = ZTMP2_DEVICE(:,:,:)*ZTMP1_DEVICE(:,:,:)
    !$acc end kernels
    CALL LES_MEAN_SUBGRID( ZTMP3_DEVICE, X_LES_RES_ddxa_W_SBG_UaRt )
    !
    !$acc kernels
    ZTMP2_DEVICE(:,:,:) = PDTH_DZ(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL MZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
    CALL LES_MEAN_SUBGRID( ZTMP3_DEVICE, X_LES_RES_ddxa_Thl_SBG_UaRt )
    !
    !$acc kernels
    ZTMP2_DEVICE(:,:,:) = PDR_DZ(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL MZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
    CALL LES_MEAN_SUBGRID( ZTMP3_DEVICE, X_LES_RES_ddxa_Rt_SBG_UaRt )
    !
    CALL MZM_DEVICE(PEMOIST,ZTMP2_DEVICE)
    !$acc kernels
    ZTMP3_DEVICE(:,:,:) = ZTMP2_DEVICE(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL MZF_DEVICE( ZTMP3_DEVICE, ZTMP4_DEVICE )
    CALL LES_MEAN_SUBGRID( ZTMP4_DEVICE, X_LES_SUBGRID_WThv , .TRUE. ) 
    !
    !$acc kernels
    ZTMP2_DEVICE(:,:,:) = -XCTP*PSQRT_TKE(:,:,:)/PLM(:,:,:)*ZTMP1_DEVICE(:,:,:)
    !$acc end kernels
    CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_SUBGRID_RtPz ) 
!$acc end data
#endif
    CALL SECOND_MNH(ZTIME2)
    XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
  END IF
!
END IF
!
!----------------------------------------------------------------------------
!
!
!*       4.   TURBULENT CORRELATIONS : <w Rc>
!             -------------------------------
!
!
!*       4.1  <w Rc>    
!
IF ( ((OTURB_FLX .AND. tpfile%lopened) .OR. LLES_CALL) .AND. (KRRL > 0) ) THEN
  !  
  ! recover the Conservative potential temperature flux : 
#ifndef MNH_OPENACC
  ZA(:,:,:)   = DZM(PIMPL * PTHLP(:,:,:) + PEXPL * PTHLM(:,:,:)) / PDZZ(:,:,:) *       &
                  (-PPHI3(:,:,:)*MZM(PLM(:,:,:)*PSQRT_TKE(:,:,:))) * XCSHF
#else
!$acc kernels present_cr(ztmp1_device)
  ZTMP1_DEVICE(:,:,:) = PIMPL * PTHLP(:,:,:) + PEXPL * PTHLM(:,:,:)
!$acc end kernels
  CALL DZM_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
!$acc kernels
  ZTMP3_DEVICE(:,:,:) = PLM(:,:,:)*PSQRT_TKE(:,:,:)
!$acc end kernels
  CALL MZM_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
!$acc kernels
  ZA(:,:,:)   = ZTMP2_DEVICE(:,:,:) / PDZZ(:,:,:) * (-PPHI3(:,:,:)*ZTMP4_DEVICE(:,:,:)) * XCSHF
#endif
  ZA(:,:,IKB) = ( PIMPL*PSFTHP(:,:) + PEXPL*PSFTHM(:,:) ) * PDIRCOSZW(:,:)
!$acc end kernels
  !  
  ! compute <w Rc>
#ifndef MNH_OPENACC
  ZFLXZ(:,:,:) = MZM( PAMOIST(:,:,:) * 2.* PSRCM(:,:,:) ) * ZFLXZ(:,:,:) + &
                 MZM( PATHETA(:,:,:) * 2.* PSRCM(:,:,:) ) * ZA(:,:,:)
#else
  !$acc kernels
  ZTMP1_DEVICE(:,:,:) = PAMOIST(:,:,:) * 2.* PSRCM(:,:,:)
  !$acc end kernels
  CALL MZM_DEVICE(ZTMP1_DEVICE, ZTMP2_DEVICE )
  !$acc kernels
  ZTMP3_DEVICE(:,:,:) =PATHETA(:,:,:) * 2.* PSRCM(:,:,:)
  !$acc end kernels
  CALL MZM_DEVICE(ZTMP3_DEVICE, ZTMP4_DEVICE )
  !$acc kernels
  ZFLXZ(:,:,:) = ZTMP2_DEVICE(:,:,:) * ZFLXZ(:,:,:) + &
                 ZTMP4_DEVICE(:,:,:) * ZA(:,:,:)
#endif
  ZFLXZ(:,:,KKA) = ZFLXZ(:,:,IKB) 
  !$acc end kernels
  !                 
  ! store the liquid water mixing ratio vertical flux
  IF ( OTURB_FLX .AND. tpfile%lopened ) THEN
    TZFIELD%CMNHNAME   = 'RCW_FLX'
    TZFIELD%CSTDNAME   = ''
    TZFIELD%CLONGNAME  = 'RCW_FLX'
    TZFIELD%CUNITS     = 'kg m s-1 kg-1'
    TZFIELD%CDIR       = 'XY'
    TZFIELD%CCOMMENT   = 'Liquid water mixing ratio vertical flux'
    TZFIELD%NGRID      = 4
    TZFIELD%NTYPE      = TYPEREAL
    TZFIELD%NDIMS      = 3
    TZFIELD%LTIMEDEP   = .TRUE.
!$acc update self(ZFLXZ)
    CALL IO_Field_write(TPFILE,TZFIELD,ZFLXZ)
  END IF
  !  
! and we store in LES configuration this subgrid flux <w'rc'>
!
  IF (LLES_CALL) THEN
    CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
    CALL LES_MEAN_SUBGRID( MZF(ZFLXZ), X_LES_SUBGRID_WRc )
#else
!$acc data copy(X_LES_SUBGRID_WRc)
    CALL MZF_DEVICE( ZFLXZ, ZTMP1_DEVICE )
    CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_SUBGRID_WRc )
!$acc end data
#endif
    CALL SECOND_MNH(ZTIME2)
    XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
  END IF
!
END IF !end of <w Rc>

IF (GOCEAN.AND.LDEEPOC) THEN
  DEALLOCATE(ZXHAT_ll,ZYHAT_ll)
END IF

if ( mppdb_initialized ) then
  !Check all inout arrays
  call Mppdb_check( pbl_depth, "Turb_ver_thermo_flux end:pbl_depth" )
  call Mppdb_check( prthls,    "Turb_ver_thermo_flux end:prthls"    )
  call Mppdb_check( prrs,      "Turb_ver_thermo_flux end:prrs"      )
  !Check all out arrays
  call Mppdb_check( pwthv, "Turb_ver_thermo_flux end:pwthv" )
  call Mppdb_check( pthlp, "Turb_ver_thermo_flux end:pthlp" )
  call Mppdb_check( prp,   "Turb_ver_thermo_flux end:prp"   )
  call Mppdb_check( ptp,   "Turb_ver_thermo_flux end:ptp"   )
  call Mppdb_check( pwth,  "Turb_ver_thermo_flux end:pwth"  )
  call Mppdb_check( pwrc,  "Turb_ver_thermo_flux end:pwrc"  )
end if

!$acc end data

#ifndef MNH_OPENACC
Deallocate( za, zflxz, zsource, zkeff, zf, zdfddtdz, zdfddrdz, z3rdmoment, zf_new, zrwthl, zrwrnp, zcld_thold, zalt )
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE()
#endif

!$acc end data

!----------------------------------------------------------------------------
END SUBROUTINE TURB_VER_THERMO_FLUX
